@@ -484,6 +484,7 @@ int readproc(int do_stat)
 	unsigned long	startcode, endcode;
 	int		pid, f;
 	ssize_t		len;
+        char            process_status[11];
 
 	/* Open the /proc directory. */
 	if (chdir("/proc") == -1) {
@@ -567,11 +568,20 @@ int readproc(int do_stat)
 
 			/* Get session, startcode, endcode. */
 			startcode = endcode = 0;
+                        /*
 			if (sscanf(q, 	"%*c %*d %*d %d %*d %*d %*u %*u "
 					"%*u %*u %*u %*u %*u %*d %*d "
 					"%*d %*d %*d %*d %*u %*u %*d "
 					"%*u %lu %lu",
 					&p->sid, &startcode, &endcode) != 3) {
+                        */
+                        if (sscanf(q,   "%10s %*d %*d %d %*d %*d %*u %*u "
+                                        "%*u %*u %*u %*u %*u %*d %*d "
+                                        "%*d %*d %*d %*d %*u %*u %*d "
+                                        "%*u %lu %lu",
+                                        process_status, 
+                                        &p->sid, &startcode, &endcode) != 4) {
+
 				p->sid = 0;
 				nsyslog(LOG_ERR, "can't read sid from %s\n",
 					path);
@@ -586,6 +596,18 @@ int readproc(int do_stat)
 			if (startcode == 0 && endcode == 0)
 				p->kernel = 1;
 			fclose(fp);
+                        if ( (strchr(process_status, 'D') != NULL) ||
+                             (strchr(process_status, 'Z') != NULL) ){
+                           /* Ignore zombie processes or processes in
+                              disk sleep, as attempts
+                              to access the stats of these will
+                              sometimes fail. */
+                              if (p->argv0) free(p->argv0);
+                              if (p->argv1) free(p->argv1);
+                              if (p->statname) free(p->statname);
+                             free(p);
+                             continue;
+                        }
 		} else {
 			/* Process disappeared.. */
 			if (p->argv0) free(p->argv0);
@@ -919,6 +941,22 @@ void usage(void)
 	exit(1);
 }
 
+
+void pidof_usage(void)
+{
+   printf("pidof usage: [options] <program-name>\n\n");
+   printf(" -c           Return PIDs with the same root directory\n");
+   printf(" -h           Display this help text\n");
+   printf(" -f <format>  Display PIDs in a given printf-style format\n");
+   printf(" -n           Avoid using stat system function on network shares\n");
+   printf(" -o <pid>     Omit results with a given PID\n");
+   printf(" -q           Quiet mode. Do not display output\n");
+   printf(" -s           Only return one PID\n");
+   printf(" -x           Return PIDs of shells running scritps with a matchign name\n");
+   printf("\n");
+}
+
+
 /* write to syslog file if not open terminal */
 #ifdef __GNUC__
 __attribute__ ((format (printf, 2, 3)))
@@ -943,6 +981,7 @@ void nsyslog(int pri, char *fmt, ...)
 #define PIDOF_SINGLE	0x01
 #define PIDOF_OMIT	0x02
 #define PIDOF_NETFS	0x04
+#define PIDOF_QUIET     0x08
 
 /*
  *	Pidof functionality.
@@ -958,6 +997,7 @@ int main_pidof(int argc, char **argv)
 	int		chroot_check = 0;
 	struct stat	st;
 	char		tmp[512];
+        char            *format = NULL;
 
 	omit = (OMIT*)0;
 	nlist = (NFS*)0;
@@ -966,7 +1006,7 @@ int main_pidof(int argc, char **argv)
 	if ((token = getenv("PIDOF_NETFS")) && (strcmp(token,"no") != 0))
 		flags |= PIDOF_NETFS;
 
-	while ((opt = getopt(argc,argv,"hco:sxn")) != EOF) switch (opt) {
+	while ((opt = getopt(argc,argv,"qhcof:sxn")) != EOF) switch (opt) {
 		case '?':
 			nsyslog(LOG_ERR,"invalid options on command line!\n");
 			closelog();
@@ -974,6 +1014,12 @@ int main_pidof(int argc, char **argv)
 		case 'c':
 			if (geteuid() == 0) chroot_check = 1;
 			break;
+                case 'h':
+                        pidof_usage();
+                        exit(0);
+                case 'f':
+                        format = optarg;
+                        break;
 		case 'o':
 			here = optarg;
 			while ((token = strsep(&here, ",;:"))) {
@@ -999,6 +1045,9 @@ int main_pidof(int argc, char **argv)
 			}
 			flags |= PIDOF_OMIT;
 			break;
+                case 'q':
+                        flags |= PIDOF_QUIET;
+                        break; 
 		case 's':
 			flags |= PIDOF_SINGLE;
 			break;
@@ -1065,15 +1114,26 @@ int main_pidof(int argc, char **argv)
 						continue;
 					}
 				}
-				if (!first)
-					printf(" ");
-				printf("%d", p->pid);
+
+				if ( ~flags & PIDOF_QUIET ) {
+                                    if (format)
+                                       printf(format, p->pid);
+                                    else
+                                    {
+					if (! first)
+						printf(" ");
+					printf("%d", p->pid);
+                                    }
+				}
 				first = 0;
 			}
 		}
 	}
 	if (!first)
+        {
+            if ( ~flags & PIDOF_QUIET ) 
 		printf("\n");
+        }
 
 	clear_mnt();
 
